/*
 Krigin
 
 Input
 -----
 - a definition space E
 - a function f on this space, i.e. a way to compute f(x) for any x of E.
 Usually thanks to a formula.
 
 Output
 ------
 - an estimation of the variogram of f, which is a curve "gamma(x,y) vs distance(x,y)"
 http://en.wikipedia.org/wiki/Variogram
 
 - a model of this variogram.
 Four possible classes of models are used, mainly of test purpose,
 with estimation of parameters by using a simplified PSO 
 (derived from Standard PSO 2007 http://www.particleswarm.info)
  For simplicity, you may use just the Beta distributions 
  http://en.wikipedia.org/wiki/Beta_distribution
  
  - an estimation of the fitness value on a random point, thanks to the
  Krigin method (that makes use of the model of the variagram that 
				 has been found), and comparison to the right value
  --------------------------------------------------------------------------------
  Copyright (C) Maurice 2009 <Maurice.Clerc@WriteMe.com>
  
  * This is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
  * Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version ( see http://www.gnu.org/licenses/).
										*/
//--------------------------------------------------------------------------------
#include "stdio.h"
#include "math.h"
#include <stdlib.h>

#define DMax 10 //  Max dimension of the search space
#define kMax 11 //  Max number classes of distance
#define modelMax 5 // Max number of models
#define NMax 402  // N known points + 2
#define zero  1.0e-30			// 1.0e-30 // To avoid numerical instabilities with some compilers

struct matrix {int size; long double m[NMax][NMax];};
struct position {int size; long double x[DMax]; double f;};
struct vector {int size; long double v[NMax];};
#include "modelOptimStruct.c"

// Subroutines
double alea (double a, double b);
double distance(struct position X1,struct position X2);
void matrixDisplay(struct matrix M);
struct matrix matrixInvert(struct matrix M);
struct vector matrixSolve(struct matrix M, struct vector V);
struct vector matrixVect(struct matrix M,struct vector V);
struct coeff modelOptim(int model,double h[],double g[],int kSup);
double perf(struct position x,int code);
int sign (double x);
void vectorDisplay(struct vector V);
#include "modelOptimSub.c"

// File(s);
FILE * f_run;
FILE * f_coeff;

// Global variables
long double E,pi; 
double hMax, gMax;
//------------------------------------------------------------------------------
int main()
{
 struct coeff c[modelMax];
 int d,D;
 struct matrix Dist;
 double distMax;
 int functionCode;
 double g[kMax];
 struct matrix Gamma;
 struct vector gammaStar;
 double h[kMax];
 double hDelta;
 int i,j,k;
 int kSup;
 struct vector lambda;
 int model,modelNb;
 int N;
 int nbK[kMax];
 double xMin,xMax;
 struct position X[NMax];
 double zStar;
 double zz;	
 
 pi=acos ((long double) -1);
 E=exp(1);
 
 f_run = fopen ("f_run.txt", "w");  
 f_coeff= fopen("f_coeff.txt","w");
 
 // Parameters
 kSup=10; 
 
 // DATA
 functionCode=104; 
 // -1 = random . The variogram must be constant (more or less)
 // 0 = Parabola/Sphere
 // 1 = Sinus
 // 2 = Rosenbrock (D>=2)
 // 3 = Rastrigin
 // 4 = Tripod (on [-100,100]^2)
 // 5 = Ackley
 // 6 = Griewank
 // 7 = KrishnaKumar (D>=2)
 // 8 =
 // 9 = Gear train on [12,60]^4
 // 99 = linear
 // 100 = Sphere CEC 2005
 // 102 = Rosenbrock CEC 2005 [-100,100], D >=2
 // 103 = Rastrigin CEC 2005
 // 104 = Schwefel (F2) CEC 2005
 // 105 = Griewank CEC 2005, but NOT rotated
 // 106 = Ackley CEC 2005  [-32,32]
 
 D=10; // Dimension
 xMin=-10; 
 xMax=10;
 // nb of points N, 
 N=33*D+67; // Just a rule of thumb
 hDelta=(xMax-xMin)*2./N;
 
 for(i=0;i<N;i++) 
 {
  X[i].size=D;
 }
 for(i=0;i<N;i++) 
 {
  for(d=0;d<D;d++)
   X[i].x[d]=alea(xMin,xMax);  
  
  X[i].f=perf(X[i],functionCode);
 }
 
 
 // Distance matrix
 Dist.size=N+1;
 for (i=0;i<N;i++)
 {
  for(j=i+1;j<N+1;j++)
  {
   Dist.m[i][j]=distance(X[i],X[j]);
   //printf("\n %f %f %f ",X[i].x[0],X[j].x[0],Dist.m[i][j]);
  }
 }
 
 // Max distance for the N points
 distMax=0;
 for (i=0;i<N-1;i++)
 {
  for (j=i+1;j<N;j++)
  {
   if(Dist.m[i][j]>distMax) distMax=Dist.m[i][j];
  }
 }
 
 printf("\ndistMax %f",distMax);
 
 // Variogram estimation 
 // For different h values (kSup+1 classes)
 //  find the n(h) couples (x_i,x_j) for which  the distance is about equal to h
 //  and compute somme((f(x_i)-f(x_j))^2)/n(h)
 
 for(k=0;k<=kSup;k++) {g[k]=0; nbK[k]=0; h[k]=k*distMax/kSup;}
 
 for (i=0;i<N-1;i++)
 {
  for (j=i+1;j<N;j++)
  {
   for(k=0;k<=kSup;k++)
   {
	if(fabs(Dist.m[i][j]-h[k])<=hDelta)
	{
	 zz=X[i].f-X[j].f;
	 g[k]=g[k]+zz*zz;
	 nbK[k]=nbK[k]+1;
	}			
   }	
  }
 }
 
 // Variogram
 for(k=0;k<=kSup;k++)
 {
  if(nbK[k]>0) 
  {
   g[k]=g[k]/nbK[k];
  }
 }
 
 for(k=0;k<=kSup;k++)
 {
  printf("\n %f %f %i",h[k],g[k],nbK[k]);
  fprintf(f_run,"%f %f \n",h[k],g[k]); // Save the variogram
 }
 
 // Variogram model, gamma (function g(h) vs h)
 //0 Power
 //1 Exponential  (very rarely the best)
 //2 "Parabolic"
 //3 Beta distribution (very often the best)
 // (See modelG.c for details)
 
 
 modelNb=4;
 
 for(model=0;model<modelNb;model++) // For each model
 {
  //  find the best parameters
  c[model]=modelOptim(model,h,g,kSup);
 }
 
 // Choose the best model
 model=0;
 for(i=1;i<modelNb;i++)
 {
  if(c[i].error<c[model].error) model=i;
 }
 
 switch(model)
 {
  case 0:
   printf("\nBest model = %i := %f*h^%f",model,c[0].c[0],c[0].c[1]);
   break;
  case 1:
   printf("\nBest model = %i := %f*(1-exp(-h/%f))",model,c[1].c[0],c[1].c[1]);
   break;
  case 2:
   printf("\nBest model = %i := %f*(1-(1-h/%f)^%f)",model,c[2].c[0],c[2].c[1],c[2].c[2]);
   break;
  case 3:
   zz=
	pow(c[3].c[0],c[3].c[0])*pow(c[3].c[1],c[3].c[1])
	*pow(hMax/(c[3].c[0]+c[3].c[1]),c[3].c[0]+c[3].c[1]);
   printf("\nBest model = %i := h^%f*(%f-h)^%e/%e",model,c[3].c[0],hMax,c[3].c[1],zz/gMax);
   break;
   
 }
 //------------------------ Now, let's try ----------------------------------
 // Point to sample (rank N+1) 
 X[N].size=D;
 for(d=0;d<D;d++)
  X[N].x[d]=alea(xMin,xMax); // Just to test
 
 // Matrix of variogram values gamma(x_i,x_j)
 Gamma.size=N+1;
 for (i=0;i<N-1;i++)
 {
  Gamma.m[i][i]=0;
  for (j=i+1;j<N;j++)
  {
   zz=Dist.m[i][j];
   Gamma.m[i][j]=modelG(model,zz,c[model]);
  }
 }
 
 // Symmetrical
 for (i=0;i<N-1;i++)
 {
  for (j=i+1;j<N;j++)
  {
   Gamma.m[j][i]=Gamma.m[i][j];
  }
 }
 Gamma.m[N][N]=0;
 
 // Complete by 1 last row and last column	
 for (i=0;i<N;i++)
 {
  Gamma.m[N][i]=1;
  Gamma.m[i][N]=1;
 }
 
 // Display the matrix
 //	matrixDisplay(Gamma);
 
 // Vecteur of gammax_i,x* 
 gammaStar.size=N+1;
 for(i=0;i<N;i++) 
 {
  zz=distance(X[i],X[N]);
  gammaStar.v[i]=modelG(model,zz,c[model]);
 }
 gammaStar.v[N]=1;
 //	printf("\ngammaStar\n");vectorDisplay(gammaStar);	
 
 // lambda weights for the linear combination	
 lambda=matrixSolve(Gamma,gammaStar);
 //	printf("\nLambda:\n"); vectorDisplay(lambda);
 
 // Estimation of f(x*)
 zStar=0;
 for(i=0;i<N;i++) zStar=zStar+lambda.v[i]*X[i].f;
 
 printf("\n\nKrigin estimation");
 printf("\nChosen point X[N]\n ");for(d=0;d<D;d++) printf("%f ",(double) X[N].x[d]);
 printf("\nEstimated value: %f \nReal value %f",zStar,perf(X[N],functionCode));
 
 return (0);
}
// ===========================================================
double alea (double a, double b) 
{				// random number (uniform distribution) in  [a b]
 double r;
 r=a+ (double) rand ()*(b-a)/RAND_MAX;
 return r; 
}

// ===========================================================
int alea_integer (int a, int b) 
{				// Integer random number in [a b]
 int ir;
 double r;
 
 r = alea (0, 1);
 ir = (int) (a + r * (b + 1 - a));
 
 if (ir > b)	ir = b;   
 return ir;  
}
// ===========================================================
int sign (double x) 
{    
 // WARNING: this is not the classical sign function
 if (x <= 0)	return -1;    
 return 1;   
}

// ===========================================================
double distance(struct position X1,struct position X2)
{
 int d;
 double dist=0;
 double zz;
 
 for(d=0;d<X1.size;d++)
 {
  zz=X1.x[d]-X2.x[d];
  dist=dist+zz*zz;	
 }
 dist=sqrt(dist);
 return dist;
}
// ===========================================================
void matrixDisplay(struct matrix M)
{
 register short i,j;
 
 printf("\n Matrix %ix%i",M.size,M.size);
 for (i=0;i<M.size;i++)
 {
  printf("\n");
  for (j=0;j<M.size;j++)
  {
   printf("%f ",(float)M.m[i][j]);
  }
 }
}
// ===========================================================
struct vector matrixSolve(struct matrix M,struct vector V)
{
 // Gauss pivot method
 
 struct vector b;
 double coef;
 int err;
 int i,j,k,l; 
 double max;
 int n;
 double pivot,s;
 double t[NMax][NMax+1];
 
 n=V.size;
 b.size=n;
 
 // Everything is put into t
 
 for(i=0;i<n;i++)
 {
  b.v[i]=V.v[i];
  for(j=0;j<n;j++)
   t[i][j]=M.m[i][j];
  
  t[i][n]=b.v[i];
 }
 
 err=1;
 k=0;
 while (err==1 && k<n-1)
 {
  max=fabs(t[k][k]);
  l=k;
  for(i=k+1;i<n;i++)
   if(max<fabs(t[i][k]))
  {
   max=fabs(t[i][k]);
   l=i;
  }
  if(max!=0)
  {
   if(l!=k)
	for(j=k;j<n+1;j++)
   {
	pivot=t[k][j];
	t[k][j]=t[l][j];
	t[l][j]=pivot;
   }
   pivot=t[k][k];
   for(i=k+1;i<n;i++)
   {
	coef=t[i][k]/pivot;
	for(j=k+1;j<n+1;j++)
	 t[i][j] -= coef*t[k][j];
   }
  }
  else err=0;
  k++;
 }
 if(t[n-1][n-1]==0) err=0;
 if(err==1)
 {
  b.v[n-1]=t[n-1][n]/t[n-1][n-1];
  for(i=n-2;i>=0;i--)
  {
   s=t[i][n];
   for(j=i+1;j<n;j++)
	s-=t[i][j]*b.v[j];
   b.v[i]=s/t[i][i];
  }
 }
 
 if(err==1)
 {
  //for(i=0;i<n;i++) b.v[i-1]=b.v[i];
  return b;
 }
 else
 {
  printf("\nERROR. Can't invert the matrix. I return the coefficient vector");
  return V;
 }	
} 
// ==========================================================
void vectorDisplay(struct vector V)
{
 int i;
 printf("\nVector length %i\n",V.size);
 for(i=0;i<V.size;i++) printf("%f ",(double)V.v[i]);
 
}
// ===========================================================
struct vector matrixVect(struct matrix M,struct vector V)
{
 // Product of a square matrix by a vector
 struct vector MV;
 register short i,j;
 
 MV.size=V.size;
 
 for(i=0;i<V.size;i++)
 {
  MV.v[i]=0;
  for(j=0;j<V.size;j++)
   MV.v[i]=MV.v[i]+M.m[i][j]*V.v[j];		
 }
 return MV;
}
// ===========================================================
#include "modelOptim.c"

// ===========================================================
double perf(struct position X, int code)
{
 int d,DD;
 double f=0;  // Set to zero just for my stupid compiler  
 int k;
 double p;
 double s11,s12,s21 ,s22 ;
 double sum1,sum2;
 double t0,t1,tt;
 double xd;
 struct position xs;
 double x1,x2,x3,x4;
#	include "cec2005data.c"
 
 xs=X;
 
 switch(code)
 {
  case -1: // Random
   f=alea(0,1);		
  break;
  
  
  case 0: // Parabola/Sphere
   f=0;
  for(d=0;d<X.size;d++) f=f+X.x[d]*X.x[d]; 
  break;
  
  case 1: // Sinus
   f=0;
  for(d=0;d<X.size;d++) f=f+sin(X.x[d]/3); 
  break;
  
  case 2:		// Rosenbrock
   f = 0;  
  t0 = xs.x[0]  + 1;	// Solution on (0,...0) when
  // offset=0
  for (d = 1; d < xs.size; d++)
  {     
   
   t1 = xs.x[d]  + 1;	      
   tt = 1 - t0;	      
   f += tt * tt;      
   tt = t1 - t0 * t0;      
   f += 100 * tt * tt;	      
   t0 = t1;    
  }  
  break;
  
  case 3:		// Rastrigin
   f=0;
  k = 10;  
  
  for (d = 0; d < xs.size; d++)    
  {     
   xd = xs.x[d];	      
   f =f+ xd * xd - k * cos (2 * pi * xd);	    
  }	  
  f =f+ xs.size * k;  
  break;
  
  case 4:		// 2D Tripod function
   // Note that there is a big discontinuity right on the solution
   // point. 	
   x1 = xs.x[0] ; 
  x2 = xs.x[1];  
  s11 = (1.0 - sign (x1)) / 2;
  s12 = (1.0 + sign (x1)) / 2; 
  s21 = (1.0 - sign (x2)) / 2;
  s22 = (1.0 + sign (x2)) / 2;
  
  //f = s21 * (fabs (x1) - x2); // Solution on (0,0)
  f = s21 * (fabs (x1) +fabs(x2+50)); // Solution on (0,-50)  
  f = f + s22 * (s11 * (1 + fabs (x1 + 50) + fabs (x2 - 50)) 
				 + s12 * (2 +fabs (x1 - 50) + fabs (x2 - 50)));	  
  break;
  
  case 5:  // Ackley
   sum1=0;
  sum2=0;
  DD=xs.size;
  for (d=0;d<DD;d++)
  {
   xd=xs.x[d];
   sum1=sum1+xd*xd;
   sum2=sum2+cos(2*pi*xd);
  }
  f=-20*exp(-0.2*sqrt(  sum1/DD  ))-exp(sum2/DD)+20+exp(1);	
  break;
  
  case 6:		// Griewank
   f = 0; 
  p = 1;
  
  for (d = 0; d < xs.size; d++)
  {      
   xd = xs.x[d];	      
   f = f + xd * xd;	      
   p = p * cos (xd / sqrt ((double) (d + 1)));	    
  } 
  f = f / 4000 - p + 1;	  
  break;
  
  case 7: // Krishnakumar
   f = 0;
  for ( d = 0; d < xs.size - 1; d++ )
  {
   f = f + sin( xs.x[d] + xs.x[d + 1] ) + sin( 2 * xs.x[d] * xs.x[d + 1] / 3 );
  }
  break;
  
  case 8:
   f=0;
     for ( d = 0; d < xs.size ; d++ )
  {
   f = f + exp(xs.x[d]);
  }
  break;
  
  case 9: // Gear train
   x1=xs.x[0]; // {12,13, ... 60}
  x2= xs.x[1];// {12,13, ... 60}
  x3= xs.x[2];// {12,13, ... 60}
  x4= xs.x[3];// {12,13, ... 60}
  
  f=1./6.931 - x1*x2/(x3*x4);
  f=f*f;
  break;
  
#include "cec2005.c"
  
  case 99: // Linear
   f=0;
  for(d=0;d<X.size;d++) f=f+X.x[d]; 
  break;	
 }
 
 return f;
}